/* database.c -- The "Database" is an environment and a set of locks */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "locksim.h"

struct Database *create_database(void) {
	struct Database *p = malloc(sizeof(struct Database));
	if (p == NULL)
		fatal_error("Memory full.");
	p->environment = create_environment();
	p->locks = NULL;
	p->nr_locks = 0;
	p->lock_table_length = 0;
	return p;
}

void destroy_database(struct Database *d) {
	// destroy_environment(d->environment); -- No! These are saved for the final report
	free(d);
}

/* Returns the lock table index for the lock, or -1 if that variable isn't locked by that transaction */
static int find_lock_place(struct Database *database, int symbol_index, struct Transaction *transaction) {
	for (int i = database->nr_locks - 1; i >= 0; --i) {
		if (database->locks[i].symbol_index == symbol_index && database->locks[i].transaction == transaction)
			return i;
	}
	return -1;
}

/* Returns the lock table index for the lock, or -1 if that variable isn't locked by that transaction */
enum lock_type find_my_lock(struct Database *database, int symbol_index, struct Transaction *transaction) {
	int my_lock_place = find_lock_place(database, symbol_index, transaction);
	if (my_lock_place == -1)
		return UNLOCKED;
	else
		return database->locks[my_lock_place].type;
}

/* Adds a new "unlocked lock" for a variable to the database, and returns its lock table index */
static int insert_lock(struct Database *database, int symbol_index, struct Transaction *transaction) {
	if (find_lock_place(database, symbol_index, transaction) != -1)
		fatal_error("Tried to insert a duplicate lock.");
	if (database->nr_locks == database->lock_table_length) {
		if (database->lock_table_length == 0)
			database->lock_table_length = 16;
		else
			database->lock_table_length *= 4;
		database->locks =
			realloc(database->locks,
			database->lock_table_length * sizeof(struct Lock));
		if (database->locks == NULL)
			fatal_error("Memory full.");
	}
	database->locks[database->nr_locks].symbol_index = symbol_index;
	database->locks[database->nr_locks].transaction = transaction;
	database->locks[database->nr_locks].type = UNLOCKED;
	return database->nr_locks++;
}

static void remove_lock(struct Database *database, int lock_place) {
	assert(lock_place >= 0 && lock_place < database->nr_locks && database->nr_locks > 0);
	database->locks[lock_place] = database->locks[--database->nr_locks];
}

/* Returns -1 if a lock was released or downgraded, +1 if created or upgraded, and 0 if no change */
/* If we have to wait for a lock, our transaction is blocked */
int lock_variable(struct Database *database, int symbol_index, struct Transaction *transaction, enum lock_type wanted_type) {

	/* First, what is the "worst" lock on this variable? */
	int my_lock_place = -1;
	enum lock_type worst_type = UNLOCKED;
	int nr_shared_locks = 0;
	for (int i = database->nr_locks - 1; i >= 0; --i) {
		assert(database->locks[i].type != UNLOCKED);
		if (database->locks[i].symbol_index == symbol_index) {
			if (database->locks[i].type == SHARED) {
				++nr_shared_locks;
			}
			if (database->locks[i].type > worst_type) {
				worst_type = database->locks[i].type;
			}
			if (database->locks[i].transaction == transaction) {
				my_lock_place = i;
			}
		}
	}

	// printf("[Worst lock: %d]\n", (int)worst_type);

	if (wanted_type == UNLOCKED) {
		/* We want to unlock */
		if (my_lock_place == -1) {
			if (verbosity > 0) {
				printf("Warning! Transaction %s tried to unlock %s, but had no lock.\n",
					get_symbol_name_unsafe(transaction->name->leaf_value),
					get_symbol_name_unsafe(symbol_index));
			}
			return 0; /* Unlocking an unlocked variable counts as a NOP */
		}
		else {
			remove_lock(database, my_lock_place);
			wakeup_others_waiting_for(symbol_index);
			return -1;
		}
	}
	else if (worst_type == UNLOCKED) {
		/* No lock on that variable. Go ahead and lock! */
		int lockplace = insert_lock(database, symbol_index, transaction);
		database->locks[lockplace].type = wanted_type;
		return 1;
	}
	else if (worst_type == SHARED && wanted_type == SHARED) {
		if (my_lock_place != -1) {
			assert(database->locks[my_lock_place].type == SHARED);
			if (verbosity > 0) {
				printf("Warning! Transaction %s already had a shared lock on %s.\n",
					get_symbol_name_unsafe(transaction->name->leaf_value),
					get_symbol_name_unsafe(symbol_index));
			}
			return 0; /* Re-locking counts as a NOP */
		}
		else {
			/* Adding a shared lock */
			int lockplace = insert_lock(database, symbol_index, transaction);
			database->locks[lockplace].type = SHARED;
			return 1;
		}
	}
	else if (worst_type == SHARED && wanted_type == EXCLUSIVE) {
		if (my_lock_place != -1 && nr_shared_locks == 1) {
			/* We, but no one else, had a shared lock, so we can upgrade */
			assert(database->locks[my_lock_place].type == SHARED);
			database->locks[my_lock_place].type = EXCLUSIVE;
			return 1;
		}
		else {
			/* Someone else had a shared lock */
			block_transaction(transaction, symbol_index);
			return 0; /* Can't upgrade to exclusive lock, so we must wait */
		}
	}
	else if (worst_type == EXCLUSIVE && wanted_type == EXCLUSIVE) {
		if (my_lock_place != -1) {
			assert(database->locks[my_lock_place].type == EXCLUSIVE);
			if (verbosity > 0) {
				printf("Warning! Transaction %s already had an exclusive lock on %s.\n",
					get_symbol_name_unsafe(transaction->name->leaf_value),
					get_symbol_name_unsafe(symbol_index));
			}
			return 0; /* Re-locking counts as a NOP */
		}
		else {
			/* Someone else already had an exclusive lock */
			block_transaction(transaction, symbol_index);
			return 0; /* Can't upgrade to exclusive lock, so we must wait */
		}
	}
	else if (worst_type == EXCLUSIVE && wanted_type == SHARED) {
		if (my_lock_place != -1) {
			/* We have the exclusive lock, and we want to downgrade to shared */
			assert(database->locks[my_lock_place].type == EXCLUSIVE);
			database->locks[my_lock_place].type = SHARED;
			wakeup_others_waiting_for(symbol_index);
			return -1;
		}
		else {
			/* Someone else had an exclusive lock */
			block_transaction(transaction, symbol_index);
			return 0; /* Can't share-lock, so we must wait */
		}
	}
	else {
		assert(0); /* This can never happen */
		return 0;
	}
} /* lock_variable */

static char *locktype2string(enum lock_type type) {
	if (type == UNLOCKED)
		return "unlocked";
	else if (type == SHARED)
		return "shared";
	else if (type == EXCLUSIVE)
		return "exclusive";
	else
		return "impossible";
}

void show_locks(struct Database *database) {
	printf("There %s %d lock%s:\n",
		database->nr_locks == 1 ? "is" : "are",
		database->nr_locks,
		database->nr_locks == 1 ? "" : "s");
	for (int i = 0; i < database->nr_locks; ++i) {
		printf("    %d. %s locked by %stransaction %s (type: %s)\n",
			i,
			get_symbol_name_unsafe(database->locks[i].symbol_index),
			database->locks[i].transaction->state == FINISHED ? "***FINISHED*** " : "",
			get_symbol_name_unsafe(database->locks[i].transaction->name->leaf_value),
			locktype2string(database->locks[i].type));
	}
}
