/* The scheduler controls the scheduling of transaction */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "locksim.h"

static struct Transaction **all_transactions = NULL;
static int nr_transactions = 0;

static struct Transaction **runnable_transactions = NULL;
static int nr_runnable_transactions = 0;

static struct Transaction **blocked_transactions = NULL;
static int nr_blocked_transactions = 0;

/* We don't need a table for finished transactions */

static int transaction_table_length = 0; /* Same for all three transaction tables */

/* Don't add the same transaction twice */
void add_transaction(struct Transaction *t) {
	assert(t->state == RUNNABLE);
	for (int i = 0; i < nr_transactions; ++i)
		assert(all_transactions[i] != t);
	if (nr_transactions == transaction_table_length) {
		if (transaction_table_length == 0)
			transaction_table_length = 16;
		else
			transaction_table_length *= 4;
		all_transactions = realloc(all_transactions, transaction_table_length * sizeof(struct Transaction*));
		runnable_transactions = realloc(runnable_transactions, transaction_table_length * sizeof(struct Transaction*));
		blocked_transactions = realloc(blocked_transactions, transaction_table_length * sizeof(struct Transaction*));
		if (all_transactions == NULL)
			fatal_error("Memory full.");
	}
	all_transactions[nr_transactions++] = t;
	runnable_transactions[nr_runnable_transactions++] = t;
}

void clear_scheduler(void) {
	for (int i = 0; i < nr_transactions; ++i)
		destroy_transaction(all_transactions[i]);
	nr_transactions = 0;
	nr_runnable_transactions = 0;
	nr_blocked_transactions = 0;
}

/* The "blocking variable" is the symbol index of a database variable that we failed to lock */
void block_transaction(struct Transaction *t, int blocking_variable) {
	assert(t->state == RUNNABLE);
	int place = 0;
	while (place < nr_runnable_transactions && runnable_transactions[place] != t)
		++place;
	assert(place <= nr_runnable_transactions);
	runnable_transactions[place] = runnable_transactions[--nr_runnable_transactions];
	assert(nr_blocked_transactions < nr_transactions);
	blocked_transactions[nr_blocked_transactions++] = t;
	t->state = BLOCKED;
	t->blocking_variable = blocking_variable;
} /* block_transaction */

static void unblock_transaction(struct Transaction *t) {
	assert(t->state == BLOCKED);
	int place = 0;
	while (place < nr_blocked_transactions && blocked_transactions[place] != t)
		++place;
	assert(place <= nr_blocked_transactions);
	blocked_transactions[place] = blocked_transactions[--nr_blocked_transactions];
	assert(nr_runnable_transactions < nr_transactions);
	runnable_transactions[nr_runnable_transactions++] = t;
	t->state = RUNNABLE;
} /* unblock_transaction */

void finish_transaction(struct Transaction *t) {
	if (t->state == RUNNABLE) {
		int place = 0;
		while (place < nr_runnable_transactions && runnable_transactions[place] != t)
			++place;
		assert(place <= nr_runnable_transactions);
		runnable_transactions[place] = runnable_transactions[--nr_runnable_transactions];
	}
	else if (t->state == BLOCKED) {
		int place = 0;
		while (place < nr_blocked_transactions && blocked_transactions[place] != t)
			++place;
		assert(place <= nr_blocked_transactions);
		blocked_transactions[place] = blocked_transactions[--nr_blocked_transactions];
	}
	else {
		assert(0); /* This can't happen */
	}
	t->state = FINISHED;
} /* finish_transaction */

/* Wake up all blocked transactions waiting for this variable */
void wakeup_others_waiting_for(int blocking_variable) {
	/* "unblock_transaction" can remove blocked transactions, so loop from the end! */
	for (int i = nr_blocked_transactions - 1; i >= 0; --i) {
		assert(blocked_transactions[i]->state == BLOCKED);
		if (blocked_transactions[i]->blocking_variable == blocking_variable)
			unblock_transaction(blocked_transactions[i]);
	}
}

struct Transaction *get_random_runnable_transaction(void) {
	// assert(nr_runnable_transactions > 0); -- No!
	if (nr_runnable_transactions == 0)
		return NULL;
	else
		return runnable_transactions[rand() % nr_runnable_transactions];
}

struct Transaction *find_transaction_by_name(int symbol_index) {
	for (int i = 0; i < nr_transactions; ++i) {
		if (all_transactions[i]->name->leaf_value == symbol_index)
			return all_transactions[i];
	}
	return NULL;
}

int get_nr_blocked_transactions(void) {
	return nr_blocked_transactions;
}

void show_transactions(void) {
	printf("%d transactions (%d runnable, %d blocked, %d finished).",
		nr_transactions, nr_runnable_transactions, nr_blocked_transactions,
		nr_transactions - nr_runnable_transactions - nr_blocked_transactions);
	if (nr_runnable_transactions + nr_blocked_transactions > 0) {
		printf(" Showing %d unfinished:\n",
			nr_runnable_transactions + nr_blocked_transactions);
		int nr_unfinished = 0;
		for (int i = 0; i < nr_transactions; ++i) {
			if (all_transactions[i]->state != FINISHED) {
				printf("    %d. ", ++nr_unfinished);
				show_transaction(all_transactions[i]);
			}
		}
	}
	else
		printf("\n");
}

void show_transactions_compact(void) {
	for (int i = 0; i < nr_transactions; ++i) {
		printf("%s", get_symbol_name_unsafe(all_transactions[i]->name->leaf_value));
		if (i < nr_transactions - 1)
			printf(", ");
	}
}

void show_scheduler_state(void) {
	printf("[Transactions: %d runnable, %d blocked, %d finished.]\n",
		nr_runnable_transactions, nr_blocked_transactions,
		nr_transactions - nr_runnable_transactions - nr_blocked_transactions);
}
